{
  Merge multiple .ccb files
}
{$I-,X+}

TYPE
  TNewSimHeader = RECORD
                    magic : ARRAY[0..7] OF CHAR; { CCbsim0^Z }
                    count : INTEGER;
                  END;

TYPE
  TSimObject = RECORD
                 name                 : ARRAY[1..20] OF CHAR;
                 damage, crew         : INTEGER;
                 id                   : INTEGER;
                 owner, picture       : INTEGER;
                 beamt                : INTEGER;
                 beamc, level         : BYTE;
                 bays                 : INTEGER;
                 torpt, ammo, torpc   : INTEGER;
                 engine, hull, shield : INTEGER;
                 fc                   : ARRAY[1..3] OF CHAR;
                 agg, mass, flags     : INTEGER;
                 flakt, flakc         : LONGINT;
               END;

TYPE
  THull = RECORD
            Name                          : ARRAY[1..30] OF CHAR;
            Pict                          : INTEGER;
            Un2                           : INTEGER;
            Tritanium,Duranium,Molybdenum : INTEGER;
            Fuel                          : INTEGER;
            Crew                          : INTEGER;
            Engines                       : INTEGER;
            Mass                          : INTEGER;
            Tech                          : INTEGER;
            Cargo                         : INTEGER;
            Fight,Torps                   : INTEGER;
            Beam                          : INTEGER;
            Cost                          : INTEGER;
          END;

CONST
  MAX_SHIPS   = 200;
  MAX_ID      = 999;
  MAX_VERSION = 3;
  MAX_HULLS   = 220;

VAR
  planet       : TSimObject;
  ships        : ARRAY[1..MAX_SHIPS] OF TSimObject;
  hulls        : ARRAY[1..MAX_HULLS] OF THull;
  num_ships    : INTEGER;
  num_hulls    : INTEGER;
  have_planet  : BOOLEAN;
  file_version : INTEGER;
  mode         : (CombineMode, ReportMode);

CONST
  record_size: ARRAY[0..MAX_VERSION] OF WORD = (53,53,57,65);

{ Add a ship to record }
PROCEDURE AddShip(CONST o : TSimObject);
VAR i : INTEGER;
BEGIN
  { validate }
  IF (o.id <= 0) OR (o.id > MAX_ID) THEN BEGIN
    Writeln('WARNING: ship has invalid Id ', o.id, ', ignoring.');
    Exit;
  END;

  { check whether there is already a ship with the same Id }
  FOR i:=1 TO num_ships DO IF ships[i].id = o.id THEN BEGIN
    ships[i] := o;
    Exit;
  END;

  { new ship }
  IF num_ships >= MAX_SHIPS THEN BEGIN
    Writeln('WARNING: dropping ship ', o.id, '; too many ships in combined file.');
    Exit;
  END;

  Inc(num_ships);
  ships[num_ships] := o;
END;

{ Add a planet }
PROCEDURE AddPlanet(CONST o : TSimObject);
BEGIN
  IF (o.id <= 0) OR (o.id > MAX_ID) THEN BEGIN
    Writeln('WARNING: planet has invalid Id ', o.id, ', ignoring.');
    Exit;
  END;
  planet := o;
  have_planet := TRUE;
END;

{ Load a file }
PROCEDURE LoadFile(CONST name : STRING);
VAR
  F            : FILE;
  h            : TNewSimHeader;
  got          : WORD;
  this_version : INTEGER;
  i            : INTEGER;
  o            : TSimObject;
BEGIN
  Assign(F, name);
  Reset(F, 1);
  IF IOResult<>0 THEN BEGIN
    Writeln('ERROR: ', name, ': Cannot open this file.');
    Halt(1);
  END;
  BlockRead(F, h, Sizeof(h), got);
  IF (got<>Sizeof(h)) OR (h.magic[0]<>'C') OR (h.magic[1]<>'C')
   OR (h.magic[2]<>'b') OR (h.magic[3]<>'s') OR (h.magic[4]<>'i')
   OR (h.magic[5]<>'m') OR (h.magic[7]<>#26)
  THEN BEGIN
    Writeln('ERROR: ', name, ': File is invalid.');
    Halt(1);
  END;
  this_version := Ord(h.magic[6]) - Ord('0');
  IF (this_version < 0) OR (this_version > MAX_VERSION) THEN BEGIN
    Writeln('ERROR: ', name, ': Unsupported version ', this_version+1, '.');
    Halt(1);
  END;
  IF this_version > file_version THEN file_version := this_version;

  IF mode=CombineMode THEN
    Writeln('Reading ', name, '...');
  FOR i := 1 TO h.count AND $7FFF DO BEGIN
    FillChar(o, Sizeof(o), 0);
    o.agg := -1;
    o.mass := 100;
    BlockRead(F, o, record_size[this_version], got);
    IF got <> record_size[this_version] THEN BEGIN
      Writeln('ERROR: File is truncated at ship slot ', i, '.');
      Halt(1);
    END;
    AddShip(o);
  END;

  IF h.count AND $8000<>0 THEN BEGIN
    FillChar(o, Sizeof(o), 0);
    o.agg := -1;
    o.mass := 100;
    BlockRead(F, o, record_size[this_version], got);
    IF got <> record_size[this_version] THEN BEGIN
      Writeln('ERROR: File is truncated at planet.');
      Halt(1);
    END;
    AddPlanet(o);
  END;

  IF IOResult<>0 THEN BEGIN
    Writeln('ERROR: I/O error.');
    Halt(1);
  END;

  Close(F);
END;

{ Save everything to a file }
PROCEDURE SaveFile(CONST name : STRING);
VAR
  F            : FILE;
  did          : WORD;
  i            : INTEGER;
CONST h : TNewSimHeader = (magic : 'CCbsim0'#26; count : 0);
BEGIN
  Assign(F, name);
  Rewrite(F, 1);
  IF IOResult<>0 THEN BEGIN
    Writeln('ERROR: ', name, ': Cannot create.');
    Halt(1);
  END;

  Writeln('Writing ', name, '...');

  h.magic[6] := Chr(Ord('0') + file_version);
  h.count := num_ships;
  IF have_planet THEN h.count := h.count OR $8000;
  BlockWrite(F, h, Sizeof(h), did);

  FOR i:=1 TO num_ships DO BEGIN
    BlockWrite(F, ships[i], record_size[file_version], did);
    IF did<>record_size[file_version] THEN BEGIN
      Writeln('ERROR: Write error - disk full?');
      Halt(1);
    END;
  END;
  IF have_planet THEN BEGIN
    BlockWrite(F, planet, record_size[file_version], did);
    IF did<>record_size[file_version] THEN BEGIN
      Writeln('ERROR: Write error - disk full?');
      Halt(1);
    END;
  END;
  Close(F);
END;

PROCEDURE LoadHulls;
VAR
  i, j : INTEGER;
  r    : WORD;
  f    : FILE;
BEGIN
  num_hulls := 0;
  Assign(f, 'hullspec.dat');
  Reset(f, 1);
  IF IOResult<>0 THEN BEGIN
    Writeln('WARNING: unable to load HULLSPEC.DAT, hulls will not be named.');
    Exit;
  END;
  WHILE num_hulls < MAX_HULLS DO BEGIN
    BlockRead(F, hulls[num_hulls+1], Sizeof(THull), r);
    IF R<>Sizeof(THull) THEN Break;
    Inc(num_hulls);
  END;
  Close(f);
END;

PROCEDURE GetStr(VAR data; len : INTEGER; VAR s : STRING);
VAR i : INTEGER;
BEGIN
  s[0] := Chr(len);
  Move(data, s[1], len);
  i := Pos(#0, s);
  IF i<>0 THEN s[0] := Chr(i-1);
END;

PROCEDURE FillStr(VAR s : STRING; len : INTEGER);
BEGIN
  WHILE Length(s) < len DO BEGIN
    Inc(s[0]);
    s[Length(s)] := ' ';
  END;
END;

PROCEDURE Report;
VAR
  i         : INTEGER;
  typ, nam  : STRING[20];
BEGIN
  Writeln('ID# Pl  Type             Name             Eng  Beams   T/F  Ammo  Dam%  Lvl');
  Writeln('--- --  ---------------  ---------------  ---  -----  ----- ----  ----  ---');
  FOR i := 1 TO num_ships DO BEGIN
    GetStr(ships[i].name, 15, nam);
    IF (ships[i].hull > 0) AND (ships[i].hull <= num_hulls) THEN BEGIN
      GetStr(hulls[ships[i].hull].name, 15, typ)
    END ELSE BEGIN
      Str(ships[i].hull, typ);
      typ := 'Hull #' + typ;
    END;
    FillStr(typ, 15);
    FillStr(nam, 15);
    Write(ships[i].id:3, ' ',
          ships[i].owner:2, '  ',
          typ, '  ',
          nam, '  ',
          ships[i].engine:3, '  ');
    IF (ships[i].beamt<>0) AND (ships[i].beamc<>0) THEN BEGIN
      Write(ships[i].beamc:2, 'x', ships[i].beamt);
      IF ships[i].beamt < 10 THEN Write(' ');
    END ELSE Write('  -  ');
    Write('  ');
    IF (ships[i].torpt<>0) AND (ships[i].torpc<>0) THEN BEGIN
      Write(ships[i].torpc:2, 'x', ships[i].torpt);
      IF ships[i].torpt < 10 THEN Write(' ');
    END ELSE IF (ships[i].bays<>0) THEN BEGIN
      Write(ships[i].bays:2, ' FB');
    END ELSE Write('  -  ');
    Writeln(' ', ships[i].ammo:4, '  ',
            ships[i].damage:3, '%  ',
            ships[i].level:2);
  END;
END;

PROCEDURE Help;
BEGIN
  Writeln;
  Writeln('Usage: MERGECCB [-c] inputfiles.... outputfile');
  Writeln('       MERGECCB -r inputfiles...');
  Writeln;
  Writeln('This program will take a set of .ccb files (PCC / CCBSim saved');
  Writeln('simulation setups) and merge them. It supports file format versions');
  Writeln('1 to 4, i.e. PCC 1.0 up to 1.1.13 (current as of April 2006).');
  Writeln;
  Writeln(' -c   Read input files and save output file with combined simulation.');
  Writeln('      This is the default.');
  Writeln(' -r   Read input files and print report to standard output. You can');
  Writeln('      redirect it to a file or printer using the ''>'' operator.');
  Writeln;
  Writeln('Send comments, bug reports and questions to <Streu@gmx.de>.');
  Halt;
END;

VAR
  i, j, n : INTEGER;
  s       : STRING[31];

BEGIN
  FileMode := 0;
  Writeln('MERGECCB v0.2 - Merge .ccb files - April 2006 by Stefan Reuther');
  n := ParamCount;
  IF n < 2 THEN Help;

  file_version := 0;
  num_ships := 0;
  mode := CombineMode;

  i := 1;
  WHILE i <= n DO BEGIN
    s := ParamStr(i);
    IF (s[1] = '/') OR (s[1] = '-') THEN BEGIN
      FOR j:=2 TO Length(s) DO CASE Upcase(s[j]) OF
        'R': mode := ReportMode;
        'C': mode := CombineMode;
        '?': Help;
        '/','-':;
      ELSE
        Writeln('MERGECCB: Unknown option ''', s[1], s[j], '''. Use ''MERGECCB -?'' for help.');
        Halt(1);
      END;
    END ELSE BEGIN
      Break;
    END;
    Inc(i);
  END;

  CASE Mode OF
    CombineMode: BEGIN
      WHILE i < n DO BEGIN
        LoadFile(ParamStr(i));
        Inc(i);
      END;
      IF (num_ships=0) AND NOT have_planet THEN BEGIN
        Writeln('ERROR: no objects in combined file.');
        Halt(1);
      END;
      SaveFile(ParamStr(n));
      Writeln('Combined file is version ', file_version+1, ' and has ', num_ships, ' ships.');
    END;
    ReportMode: BEGIN
      LoadHulls;
      WHILE i <= n DO BEGIN
        LoadFile(ParamStr(i));
        Inc(i);
      END;
      Report;
    END;
  END;
END.
